;===============================================================;
;                                                               ;
; Alien Breed 5                                                 ;
; Collision checking routines                                   ;
;                                                               ;
;===============================================================;

;------------------------------------------------
; checkBulletEnemyCollisions - check for collisions between bullets & enemies
;   input:  none
;   output: none
;------------------------------------------------
checkBulletEnemyCollisions:
        di                                      ; disable interrupts as IY will be used
        push    iy                              ; save IY (system flag ptr)
        ld      iy,bulletTable                  ; IY = bullet data
        ld      b,NUM_BULLETS                   ; B = number of bullets to check
cbecLoop:
        push    bc                              ; save bullet entry counter
        ld      a,(iy+0)
        or      a                               ; bullet in this entry?
        jr      z,endCBECLoop                   ; if not, go to next entry
        call    loadBulletCollideObject1        ; A = damage value; bullet loaded to collideObject1
        ld      (cbecDamage),a                  ; save damage value to code for later
        ld      ix,enemyTable                   ; IX => enemy data
        ld      b,NUM_ENEMIES                   ; B = number of enemies to check
cbecLoop2:
        push    bc                              ; save enemy entry counter
        ld      a,(ix+0)
        or      a                               ; enemy in this entry?
        jr      z,endCBECLoop2                  ; if not, go to next entry
        call    loadEnemyCollideObject2         ; load enemy into collideObject2
        call    checkCollision                  ; check for collision
        jr      nc,endCBECLoop2                 ; if no collision, go to next entry
        xor     a
        ld      (aflagAliens),a                 ; reset flag, denoting that there have been aliens shot during this level
        push    ix                              ; save enemyTable ptr
        push    iy
        pop     ix                              ; IX => bulletTable ptr
        call    removeBulletCheckAnim           ; check if bullet requires an explosion animation & remove it (located in shoot.asm)
        pop     ix                              ; restore enemyTable ptr
cbecAfterEndBullet:
        ld      a,(ix+7)                        ; A = enemy health
        ld      de,(cbecDamage)                 ; E = (cbecDamage)
        sub     e
        ld      (ix+7),a                        ; save new health value
        jr      nc,endCBECLoop2                 ; if health didn't go down past 0, go to next entry
        call    enemyDead
endCBECLoop2:
        pop     bc                              ; restore enemy entry counter
        ld      de,ENEMY_SIZE
        add     ix,de                           ; IX => next enemy entry
        djnz    cbecLoop2
endCBECLoop:
        pop     bc                              ; restore bullet entry counter
        ld      de,BULLET_SIZE
        add     iy,de                           ; IY => next bullet entry
        djnz    cbecLoop
        pop     iy                              ; restore system flag ptr
        ei                                      ; thus, interrupts can be enabled again
        ret

;------------------------------------------------
; checkBulletBossCollisions - check for collisions between bullets & boss
;   input:  none
;   output: none
;------------------------------------------------
checkBulletBossCollisions:
        xor     a
        ld      (bossShot),a                    ; clear flag
        ld      hl,collideObject2
        ld      de,(bossX)
        ld      (hl),e                          ; load x coord LSB
        inc     hl
        ld      (hl),d                          ; load x MSB
        inc     hl
        ld      (hl),16                         ; load width
        inc     hl
        ld      de,(bossY)
        ld      (hl),e                          ; load y coord LSB
        inc     hl
        ld      (hl),d                          ; load y MSB
        inc     hl
        ld      (hl),24                         ; load height
        di                                      ; no interrupts so we can use IY
        push    iy                              ; save system flag ptr
        ld      iy,bulletTable                  ; IY => start of bullet data
        ld      b,NUM_BULLETS                   ; B = number of entries to check
cbbcLoop:
        push    bc                              ; save loop counter
        ld      a,(iy+0)
        or      a                               ; is there a bullet in this entry?
        jr      z,endCBBCLoop                   ; if not, skip this entry
        call    loadBulletCollideObject1        ; A = damage value; bullet loaded to collideObject1
        ld      (cbbcDamage),a                  ; save damage value to code for later
        call    checkCollision                  ; check for a collision
        jr      nc,endCBBCLoop                  ; if no collision, next entry
        ld      a,1
        ld      (bossShot),a                    ; set flag
        ld      (iy+0),0                        ; clear entry
        ld      hl,(bossHealth)
#ifdef TI83PLUS
        ld      de,(cbbcDamage)
        ld      d,0                             ; DE = (cbbcDamage)
#endif
#ifdef TI84PLUSCE
        ld      de,0
        ld      a,(cbbcDamage)
        ld      e,a
#endif
        or      a                               ; clear carry
        sbc     hl,de                           ; HL = boss health after taking damage
        ld      (bossHealth),hl                 ; save it
        jp      c,bossDead                      ; if it went below 0, boss is dead; no more collision checking req'd
endCBBCLoop:
        pop     bc                              ; restore loop counter
        ld      de,BULLET_SIZE
        add     iy,de                           ; IY => next entry
        djnz    cbbcLoop
        pop     iy                              ; restore system flag ptr
        ei                                      ; enable interrupts again
        ret

;------------------------------------------------
; checkP1EnemyCollisions - check player 1 & enemy collisions (including enemy bullet)
;   input:  none
;   output: none
;------------------------------------------------
checkP1EnemyCollisions:
        ld      hl,p1x
        jr      checkPlayerEnemyCollisions

;------------------------------------------------
; checkP2EnemyCollisions - check player 2 & enemy collisions (including enemy bullet)
;   input:  none
;   output: none
;------------------------------------------------
checkP2EnemyCollisions:
        ld      hl,p2x
        ; fall through to checkPlayerEnemyCollisions

;------------------------------------------------
; checkPlayerEnemyCollisions - check for collisions between player & enemies
;   input:  HL => player data
;   output: none
;------------------------------------------------
checkPlayerEnemyCollisions:
        ld      (pPlayer),hl
        call    loadPlayerCollideObject1
        ld      ix,enemyTable                   ; IX => start of enemy data
        ld      b,NUM_ENEMIES                   ; B = number of enemies to check
cpecLoop:
        push    bc                              ; save counter
        ld      a,(ix+0)
        or      a                               ; enemy in this entry?
        jr      z,endCPECLoop                   ; if not, skip this entry
        call    loadEnemyCollideObject2
        call    checkCollision                  ; check for collision
        jr      nc,endCPECLoop                  ; if no collision, skip this entry
        ld      hl,(pPlayer)
        ld      de,PI
        add     hl,de
        ld      a,(hl)
        inc     hl
        or      (hl)                            ; check if player dead or invincible
        jr      nz,cpecAfterHurtPlayer          ; if one of the above, skip hurting the player
#ifdef TI83PLUS
        ld      h,0
#endif
#ifdef TI84PLUSCE
        or a \ sbc hl,hl
#endif
        ld      l,(ix+0)
        dec     l                               ; HL = enemy type (0->..)
        ld      de,enemyDamageTable
        add     hl,de                           ; HL => enemy damage
        ld      d,(hl)                          ; D = damage to take on player
        ld      a,(e1cheatFlags)
        bit     insaneModeCheat,a
        jr      z,cpecAfterInsane
        sla     d
        sla     d                               ; x4 damage for INSANE
        jr      cpecAfterTough
cpecAfterInsane:
        ld      a,(difficulty)
        or      a                               ; check difficulty level
        jr      z,cpecAfterTough                ; if NORMAL, don't double damage
        sla     d                               ; double the damage for TOUGH
cpecAfterTough:
        ld      hl,(pPlayer)
        ld      bc,PH
        add     hl,bc                           ; HL => player health
#ifndef INVINCIBLE
        ld      a,(hl)
        sub     d                               ; A = new player health after being hit by enemy
        ld      (hl),a                          ; save it
        ld      bc,PHT-PH
        add     hl,bc                           ; HL => player hurt flag
        ld      (hl),1                          ; set flag for player getting hurt this frame
#endif
cpecAfterHurtPlayer:
        call    enemyDead                       ; kill enemy and start an explosion animation
endCPECLoop:
        pop     bc                              ; restore counter
        ld      de,ENEMY_SIZE   
        add     ix,de                           ; IX => next entry in enemyTable
        djnz    cpecLoop
        ; fall through to checkPlayerEbulletCollisions

;------------------------------------------------
; checkPlayerEbulletCollisions - check for collisions between player and enemy bullet
;   input:  none
;   output: none
;   notes:  must be run directly after checkPlayerEnemyCollisions so that player is still loaded in collideObject1
;------------------------------------------------
checkPlayerEbulletCollisions:
        ld      a,(ebullet)
        or      a
        ret     z                               ; leave if there is no enemy bullet active
        ld      hl,collideObject2
        ld      de,(ebullet+2)                  ; DE = x coord
        ld      (hl),e
        inc     hl
        ld      (hl),d
        inc     hl
        ld      (hl),3                          ; currently all enemy bullets are fixed at 3x3
        inc     hl
        ld      de,(ebullet+4)                  ; DE = y coord
        ld      (hl),e
        inc     hl
        ld      (hl),d
        inc     hl
        ld      (hl),3
        call    checkCollision
        ret     nc                              ; leave if not colliding
        ld      hl,(pPlayer)
        push    hl
        ld      de,PI
        add     hl,de
        ld      a,(hl)
        inc     hl
        or      (hl)                            ; check if player dead or invincible
        jr      nz,cpebcAfterHurtPlayer         ; if one of the above, skip hurting the player
        ld      d,EBULLET_DAMAGE                ; D = base damage
        ld      a,(e1cheatFlags)
        bit     insaneModeCheat,a
        jr      z,cpebcAfterInsane              ; if insane mode not enabled, skip it
        sla     d
        sla     d                               ; x4 damage for INSANE
        jr      cpebcAfterTough
cpebcAfterInsane:
        ld      a,(difficulty)
        or      a
        jr      z,cpebcAfterTough               ; if NORMAL, just use base damage
        sla     d                               ; x2 damage for TOUGH
cpebcAfterTough:
        pop     hl
        ld      bc,PH
        add     hl,bc                           ; HL => player health
#ifndef INVINCIBLE
        ld      a,(hl)
        sub     d
        ld      (hl),a                          ; save new player health after damage taken
        ld      bc,PHT-PH
        add     hl,bc
        ld      (hl),1                          ; set player hurt flag
#endif
cpebcAfterHurtPlayer:
        xor     a
        ld      (ebullet),a
        ret

;------------------------------------------------
; checkP1BossCollisions - check player 1 & boss collisions
;   input:  none
;   output: none
;------------------------------------------------
checkP1BossCollisions:
        ld      hl,p1x
        jr      checkPlayerBossCollisions

;------------------------------------------------
; checkP2BossCollisions - check player 2 & boss collisions
;   input:  none
;   output: none
;------------------------------------------------
checkP2BossCollisions:
        ld      hl,p2x
        ; fall through to checkPlayerBossCollisions

;------------------------------------------------
; checkPlayerBossCollisions - check for collisions between player & boss
;   input:  HL => player data
;   output: none
;   notes:  must be run directly after checkBulletBossCollisions so that boss is still loaded in collideObject2
;------------------------------------------------
checkPlayerBossCollisions:
        ld      (pPlayer),hl
        ld      de,PI
        add     hl,de
        ld      a,(hl)
        inc     hl
        or      (hl)
        ret     nz                              ; if player is dead or invincible, leave
        call    loadPlayerCollideObject1
        call    checkCollision
        ret     nc                              ; if not colliding, leave
        ld      hl,(pPlayer)
        ld      bc,PH
        add     hl,bc                           ; HL => player health
        ld      (hl),-1                         ; set player health to -1 (player is killed)
        ld      bc,PHT-PH
        add     hl,bc
        ld      (hl),1                          ; set flag to say player was hurt this frame
        ret

;------------------------------------------------
; checkEnemyCollisions - check for collisions between enemies
;   input:  (enemyXTest) = x coord of test enemy
;           (enemyYTest) = y coord of test enemy
;           (enemyCollCheck) = value to use to check if too many collisions
;   output: CA = enemy not colliding
;------------------------------------------------
checkEnemyCollisions:
        push    ix                              ; IX needs to be saved for outside routines
        ld      ix,enemyTable                   ; IX = enemy data
        ld      bc,NUM_ENEMIES*256+0            ; B = number of entries to check; C = 0 (collision counter)
cecLoop:
        ld      a,(ix+0)
        or      a                               ; 0?
        jr      z,endCECLoop                    ; if so, no enemy in this entry
        ld      hl,(enemyXTest)
#ifdef TI84PLUSCE
        ld      de,0
#endif
        ld      e,(ix+2)
        ld      d,(ix+3)                        ; DE = enemy x
        or      a
        sbc     hl,de
        ld      a,l                             ; A = x difference between enemies
        jr      nc,cecXAfterNeg                 ; if NC, the difference is already a positive value
        neg                                     ; negate to change difference to a positive value
cecXAfterNeg:
        cp      8                               ; is x difference greater or equal to 8?
        jr      nc,endCECLoop                   ; if so, go to next enemy
        ld      hl,(enemyYTest)
#ifdef TI84PLUSCE
        ld      de,0
#endif
        ld      e,(ix+4)
        ld      d,(ix+5)                        ; DE = enemy y
        or      a
        sbc     hl,de
        ld      a,l                             ; A = y difference between enemies
        jr      nc,cecYAfterNeg                 ; if NC, the difference is already a positive value
        neg                                     ; negatte to change difference to a positive value
cecYAfterNeg:
        cp      8                               ; is y difference greater or equal to 8?
        jr      nc,endCECLoop                   ; if so, go to next enemy
        inc     c                               ; enemies are colliding, to inc collision counter
endCECLoop:
        ld      de,ENEMY_SIZE
        add     ix,de                           ; IX => next enemy entry
        djnz    cecLoop                         ; loop
        ld      a,c                             ; A = collision counter
        ld      hl,enemyCollCheck
        cp      (hl)                            ; check how many enemies these coords intersected with
        pop     ix                              ; restore IX for outside routines
        ret

;------------------------------------------------
; loadBulletCollideObject1 - load a bullet to collideObject1 and get it's damage value
;   input:  IY => entry in bulletTable
;   output: A = damage value
;           IY => entry in bullet table
;------------------------------------------------
loadBulletCollideObject1:
#ifdef TI84PLUSCE
        or a \ sbc hl,hl
        push hl \ pop bc
#endif
        ld      l,(iy+0)
        dec     l
        ld      c,l                             ; C = bullet type (0->..)
#ifdef TI83PLUS
        ld      h,0                             ; HL = bullet type (0->..)
        ld      b,0                             ; BC = bullet type (0->..) for later
        add     hl,hl
        add     hl,hl
        add     hl,hl                           ; *8
#endif
#ifdef TI84PLUSCE
        ld      h,8
        mlt     hl
#endif
        ld      de,bulletSizeTable
        add     hl,de                           ; HL => size data for bullet type
        ld      a,(iy+1)
        add     a,a                             ; A = direction*2
#ifdef TI83PLUS
        ld      d,0
#endif
#ifdef TI84PLUSCE
        ld      de,0
#endif
        ld      e,a
        add     hl,de                           ; HL => bullet width
        ld      de,collideObject1
        ld      a,(iy+2)
        ld      (de),a                          ; save x LSB
        inc     de
        ld      a,(iy+3)
        ld      (de),a                          ; save x MSB
        inc     de
        ld      a,(hl)
        ld      (de),a                          ; save width
        inc     de
        inc     hl                              ; DE => bullet height
        ld      a,(iy+4)
        ld      (de),a                          ; save y LSB
        inc     de
        ld      a,(iy+5)
        ld      (de),a                          ; save y MSB
        inc     de
        ld      a,(hl)
        ld      (de),a                          ; save height
        ld      hl,bulletDamageTable
        add     hl,bc                           ; HL => bullet damage (BC was bullet type from earlier)
        ld      a,(hl)                          ; A = bullet damage
        ret

;------------------------------------------------
; loadPlayerCollideObject1 - load player to collideObject1
;   input:  (pPlayer) => player data
;   output: none
;------------------------------------------------
loadPlayerCollideObject1:
        ld      ix,(pPlayer)
        ld      hl,collideObject1
        ; fall through to loadPlayerCollideObjectHL

;------------------------------------------------
; loadPlayerCollideObjectHL - load a player to one of the collide objects
;   input:  IX => player data
;           HL => collideObject1/collideObject2
;------------------------------------------------
loadPlayerCollideObjectHL:
        ld      e,(ix+PX)
        ld      d,(ix+PX+1)
        ld      (hl),e                          ; load x LSB
        inc     hl
        ld      (hl),d                          ; load x MSB
        inc     hl
        ld      (hl),8                          ; load width
        inc     hl
        ld      e,(ix+PY)
        ld      d,(ix+PY+1)
        ld      (hl),e                          ; load y LSB
        inc     hl
        ld      (hl),d                          ; load y MSB
        inc     hl
        ld      (hl),8                          ; load height
        ret

;------------------------------------------------
; loadScreenCollideObjectIX - load screen plus border to collideObject1 or collideObject2
;   input:  IX => either collideObject1 or collideObject2
;   output: none
;------------------------------------------------
loadScreenCollideObjectIX:
        ld      hl,(screenX)
        ld      de,8
        or      a
        sbc     hl,de                           ; HL = screen x less 8
        jr      nc,loadScreenColX
        ld      hl,0                            ; can't be less than zero
loadScreenColX:
        ld      (ix),l
        ld      (ix+1),h                        ; load x
        ld      (ix+2),(TILE_COLS*8)+16         ; load width
        ld      hl,(screenY)
        or      a
        sbc     hl,de
        jr      nc,loadScreenColY
        ld      hl,0
loadScreenColY:
        ld      (ix+3),l
        ld      (ix+4),h                        ; load y
        ld      (ix+5),(TILE_ROWS*8)+16         ; load height
        ret

;------------------------------------------------
; loadScreenCollideObjectIXNoBorder - load screen to collideObject1 or collideObject2
;   input:  IX => either collideObject1 or collideObject2
;   output: none
;------------------------------------------------
loadScreenCollideObjectIXNoBorder:
        ld      hl,(screenX)
        ld      (ix),l
        ld      (ix+1),h                        ; load x
        ld      (ix+2),TILE_COLS*8              ; load width
        ld      hl,(screenY)
        ld      (ix+3),l
        ld      (ix+4),h                        ; load y
        ld      (ix+5),TILE_ROWS*8              ; load height
        ret

;------------------------------------------------
; loadEnemyCollideObject2 - load an enemy into collideObject2
;   input:  IX => entry in enemyTable
;   output: IX => entry in enemyTable
;------------------------------------------------
loadEnemyCollideObject2:
        ld      hl,collideObject2
        ld      a,(ix+2)
        ld      (hl),a                          ; save x LSB
        inc     hl
        ld      a,(ix+3)
        ld      (hl),a                          ; save x MSB
        inc     hl
        ld      (hl),8                          ; save width
        inc     hl
        ld      a,(ix+4)
        ld      (hl),a                          ; save y LSB
        inc     hl
        ld      a,(ix+5)
        ld      (hl),a                          ; save y MSB
        inc     hl
        ld      (hl),8                          ; save height
        ret

;------------------------------------------------
; loadObjectCollideObject2 - load an object into collideObject2
;   input:  IX => entry in objectTable
;   output: IX => entry in objectTable
;------------------------------------------------
loadObjectCollideObject2:
        ld      hl,collideObject2
        ld      a,(ix+1)
        ld      (hl),a
        inc     hl
        ld      a,(ix+2)
        ld      (hl),a
        inc     hl
        ld      (hl),8
        inc     hl
        ld      a,(ix+3)
        ld      (hl),a
        inc     hl
        ld      a,(ix+4)
        ld      (hl),a
        inc     hl
        ld      (hl),8
        ret

;------------------------------------------------
; checkCollision - check for collisions between two objects
; modified version of a similar routine by Patrick Davidson
;   input:  (collideObject1) = object 1 [x{2}, width, y{2}, height]
;           (collideObject2) = object 2 [x{2}, width, y{2}, height]
;   output: CA = 1 if they are colliding
;------------------------------------------------
checkCollision:
        ld      hl,(collideObject1)             ; HL = obj1 x
        ld      de,(collideObject2)             ; DE = obj2 x
#ifdef TI84PLUSCE
        call    _SetHLUto0
        call    _SetDEUto0
#endif
        or      a                               ; clear carry
        sbc     hl,de
        jr      c,checkCollision1               ; if negative, obj1 is to the left of obj2, so check difference against obj1 width
        ld      a,(collideObject2+2)
        ld de,0 \ ld e,a                        ; DE = obj2 width; HL = difference; CA = 0
        sbc     hl,de                           ; check if difference is greater than obj2 width
        ret     nc                              ; if so, no collision, return with CA = 0
        jr      checkCollision2
checkCollision1:
        call    absHL                           ; HL = abs(HL)
        ld      a,(collideObject1+2)
        ld de,0 \ ld e,a                        ; DE = obj1 width
        or      a                               ; clear carry
        sbc     hl,de                           ; check if difference is greater than obj1 width
        ret     nc                              ; if so, no collision, return with CA = 0
checkCollision2:
        ld      hl,(collideObject1+3)           ; HL = obj1 y
        ld      de,(collideObject2+3)           ; DE = obj2 y
#ifdef TI84PLUSCE
        call    _SetHLUto0
        call    _SetDEUto0
#endif
        or      a                               ; clear carry
        sbc     hl,de
        jr      c,checkCollision3               ; if negative, obj1 is above obj2, so check difference against obj1 height
        ld      a,(collideObject2+5)
        ld de,0 \ ld e,a                        ; DE = obj2 height; HL = difference; CA = 0
        sbc     hl,de                           ; check if difference is greater than ob2 height
        ret                                     ; ret regardless with the carry flag result
checkCollision3:
        call    absHL                           ; HL = abs(HL)
        ld      a,(collideObject1+5)
        ld de,0 \ ld e,a                        ; DE = obj1 height
        or      a                               ; clear carry
        sbc     hl,de                           ; check if difference is greater than obj1 height
        ret                                     ; ret regardless with the carry flag result

.end
